#ifndef M_STRUCTS_H
#define M_STRUCTS_H
/* Pxys - An IRC ProxyScan, m_structs.h
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * $Id: m_structs.h,v 1.29 2002/12/16 01:10:17 mbuna Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#include <sys/types.h>          /* time_t */
#include <netinet/in.h>         /* in_addr */

#include "numnicks.h"
#include "typedefs.h"
#include "pxys_sem.h"

#ifdef ENABLE_POLL_EMULATION
#include "poll.h"
#endif

#if !defined(ENABLE_POLL_EMULATION) && defined(HAVE_SYS_POLL_H)
#include <sys/poll.h>
#endif


/* state defines of a scan thread */
#define STH_STATE_RUNNING			0x0001  /* Thread is scanning IPs. */
#define STH_STATE_WAITING			0x0002  /* This thread has nothing to scan. */

#define scanth_isrunning(t) ((t)->state & STH_STATE_RUNNING)
#define scanth_iswaiting(t) ((t)->state & STH_STATE_WAITING)
#define scanth_setrunning(t) (t)->state = STH_STATE_RUNNING
#define scanth_setwaiting(t) (t)->state = STH_STATE_WAITING


#define SCANTHREAD_CHECKING_WINGATE     0x0008
#define SCANTHREAD_CHECKING_SOCKS       0x0010
#define SCANTHREAD_CHECKING_HTTP        0x0020

#define SCANTHREAD_RESCAN           0x0100
#define SCANTHREAD_TIMEWAIT         0x0200

#define SIT_INACTIVE                0x0000
#define SIT_ACTIVE                  0x0001
#define SIT_TIMEOUT                 0x0002
#define SIT_ERROR                   0x0004
#define SIT_INPROGRESS              0x0008
#define SIT_RECEIVED                0x0010
#define SIT_FOUND                   0x0020
#define SIT_NOTFOUND                0x0040
#define SIT_NOSOCKS                 0x0080
#define SIT_UNREACH                 0x0100

#define SIT_CONNECTED               0x1000

/* Be sure scanItem is connected... */
#define IsConnecting(s) (!(s->state & SIT_CONNECTED))
#define SetConnected(s) s->state |= SIT_CONNECTED;

#ifndef SCANS_PER_THREAD
#warning SCANS_PER_THREAD not defined, using default value 10
#define SCANS_PER_THREAD            10
#endif

#ifndef NICKLEN
#define NICKLEN                     9 /* RFC? */
#endif
#define FLAG_NULL                   0x00000000
#define FLAG_IDENT                  0x00000001  /* User is idented (no ~) */
#define FLAG_OPER                   0x00000002  /* Global operator (+o) */

#define FLAG_SHOWOPR                0x00010000  /* showopers feature */


/* All (remote) IRC clients are stored in an array of struct Client */
struct Client
{
   unsigned int nserv;
   unsigned int nnick;
   char nick[NICKLEN+1];        /* auto nick[10] */ 
   unsigned int flags;          /* Flags: operator? ident'd?... */
   struct in_addr ip;           /* Real IP */
   time_t last;                 /* Last scanned */
};

/* The servers... */
struct Server
{
   void** cliH;                 /* Link to the clients table */
   
   char* name;                  /* Server name */
   unsigned int myx;            /* Server numeric */
   unsigned int upx;            /* Uplink numeric */
   unsigned int nn_mask;        /* Numeric nick mask */
   unsigned int clients;        /* Number of clients */
   int noscan;                  /* NoScan toggle */
   int shared;                  /* Shared toggle */
   int bursting;                /* Server is bursting? */
   unsigned int maxclients;     /* Max clients record */
   unsigned int currtoasted;    /* Toasted clients (proxies found on this server) */
   unsigned int lasttoasted;    /* Recorded value */
};


struct scanItem {
   unsigned int nserv;
   unsigned int nnick;
   struct in_addr addr;     
   unsigned short state;
   time_t timeout;
   struct sockaddr_in clientAddr;
   int tryread;
   int retry;                     /* Retry counter if EINTR or EPIPE during socket
                                   * operation
                                   */
   int dont_cache;                /* Host unreachable? We do not want it in the IP cache! */
};


struct scanthread_s
{
   pthread_t threadID;            /* single thread ID */
   int counter;                   /* number of SIT_ACTIVE items */
   struct scanItem item[SCANS_PER_THREAD];
   struct pollfd poll_fds[SCANS_PER_THREAD];
   unsigned int state;            /* Customized scanthread state */
   unsigned int port;             /* HTTP port being scanned, if any */
   pxys_sem_t pxySem;             /* semaphore used by the listener to tell the appropriate
                                   * thread that the proxy has replied (vulnerable CONNECT..)
                                   */
   struct scanthread_s *next;     /* the next thread info */
};


typedef struct scanthread_s scanthread_s;


struct scanstats_s {
   int running_threads;           /* Number of running threads */
   int waiting_threads;           /* Number of waiting threads */
   int socks5_gssapi_count;       /* Stats: GSSAPI socks5 */
   int socks5_userpw_count;       /* Stats: USER/PW socks5 */
   int socks5_notacc_count;       /* Stats: Rejected socks5 (not acceptable method) */
};

typedef struct scanstats_s scanstats_s;

/* --------------------------- Conf structs --------------------------- */

/* Sizes */
#define MAXSERVERNAME_LENGTH	64
#define MAXSERVERINFO_LENGTH	64
#define MAXPASSWORD_LENGTH		32
#define MAXNICKNAME_LENGTH		10	/* 9+1 */
#define MAXUSERNAME_LENGTH		11	/* 10+1 */
#define MAXREALNAME_LENGTH		64
#define MAXFILENAME_LENGTH		32

/* me {} */
struct Me_s {
	char name[MAXSERVERNAME_LENGTH];
	long numeric;
	long capacity;
	char serverinfo[MAXSERVERINFO_LENGTH];
	char hostip[16];
	Boolean vhost;     /* vhost for the scanner? */
	char bindip[16];
	unsigned long bindaddr;
	char listener[16];
	unsigned long listeneraddr;
	char id[10];
};

typedef struct Me_s 	Me_s;


/* uplink {} */
struct Uplink_s {
	char host[MAXSERVERNAME_LENGTH];
	Boolean vhost;     /* vhost for the hub connection? */
	char bindip[16];
	unsigned long bindaddr;
	unsigned short port;
	char password[MAXPASSWORD_LENGTH];
	int priority;
	struct Uplink_s* next;		/* Several uplink servers supported. */
};

typedef struct Uplink_s 	Uplink_s;


/* console {} */
struct Console_s {
	char * channel;
	char mode[16];
};								/* Several console channels not supported. */

typedef struct Console_s 	Console_s;


/* client {} */
struct MyClient_s {
	char nickname[MAXNICKNAME_LENGTH];
	char username[MAXUSERNAME_LENGTH];
	char * hostname;
	char * realname;
	char mode[16];
	struct in_addr ip;
	Boolean join;
	char nick[MAXNICKNAME_LENGTH];
};	

typedef struct MyClient_s 	MyClient_s;


/* perform {} */
struct Perform_s {
	long scanthreads;
	long recheck;
	long recheckshared;
	long timeout_connect;
	long timeout_send;
	long timeout_listen;
	long cachesize;
	long http_ports_count;
	unsigned short *http_ports;
};

typedef struct Perform_s 	Perform_s;


/* share {} */
struct Share_s {
   char name[MAXSERVERNAME_LENGTH];
   char successor[MAXSERVERNAME_LENGTH];
   struct Serverlist* servlist;
   struct Share_s *next;            /* More than 2 proxyscan sharing supported. */
};

typedef struct Share_s 	Share_s;


/* noscan {} */
struct NoScan_s {
	struct IPrangelist *		ips;
	struct Serverlist *			residences;
};

typedef struct NoScan_s 	NoScan_s;


struct Log_s {
    int update;
    char dir[MAXFILENAME_LENGTH];
    char history_dir[MAXFILENAME_LENGTH];
    
	char current[MAXFILENAME_LENGTH];
	char daily[MAXFILENAME_LENGTH];
	
	char pxys[MAXFILENAME_LENGTH];
	char glines[MAXFILENAME_LENGTH];
	char privmsg[MAXFILENAME_LENGTH];
	char opercmds[MAXFILENAME_LENGTH];
	char listener[MAXFILENAME_LENGTH];
	char stats[MAXFILENAME_LENGTH];
};

typedef struct Log_s 	Log_s;



/* Used for socks4 check: Some target server to test the insecurity */
struct Target_s {
	char name[MAXSERVERNAME_LENGTH];
	struct in_addr addr;
	struct Target_s * next;
};

typedef struct Target_s 	Target_s;


struct Gline_s {
	int direct;
	time_t delay;
	time_t cached;
#ifdef SCAN_WINGATE
	char reason_wingate[160];
#endif
#ifdef SCAN_SOCKS
	char reason_socks[160];
#endif
	char reason_http[160];
	struct Relay_s *relay;
};

typedef struct Gline_s 	Gline_s;


/* Sub-structs */

struct Serverlist {
	char name[MAXSERVERNAME_LENGTH];
	struct Serverlist* next;
};

typedef struct Serverlist 	Serverlist;


struct IPrangelist {
	struct in_addr 		start;
	struct in_addr 		stop;
	struct IPrangelist* next;
};

typedef struct IPrangelist 	IPrangelist;


struct Relay_s {
	char nick[NICKLEN+1];
	char serv[MAXSERVERNAME_LENGTH];
        char auth[64];
        Boolean isauth;
	struct Relay_s * next;
};

typedef struct Relay_s 	Relay_s;


#endif  /* M_STRUCTS_H */
