/* Pxys - An IRC ProxyScan, ms_msgout.c
 * 
 * Copyright (C) 2001,2002 Stephane Thiell <mbuna@undernet.org>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 * 
 * Outgoing messages formating.
 *
 * $Id: ms_msgout.c,v 1.23 2002/11/15 19:50:57 mbuna Exp $
 */


#define GLINE_SYNTAX "gline *@%s %ld %s"

#define MAX_SENDQ_ITEMS 256
#define MAX_SENDQ_BYTES MAX_SENDQ_ITEMS * sizeof(msgoutQItem)
/* 64 - 66816 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>             /* close */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif

#include "ms_globals.h"
#include "ms_msgout.h"
#include "ms_myclient.h"
#include "ms_netutils.h"

#include "typedefs.h"
#include "conf.h"
#include "misc.h"
#include "debug.h"

#include "hashtab.h"
#include "standard.h"    /* min() */
#include "m_array.h"     /* array_countclient_from_ip */
#include "m_log.h"
#include "m_opercmds.h"
#include "m_proxycache.h"
#include "m_shutdown.h"
#include "m_structs.h"
#include "m_stats.h"     /* pxys_stats */

/* xmalloc at the bottom of include files */
#include "xmalloc.h"

extern htab *mHS;
extern const char MyClientXXX[];
extern long totalfound;
extern statistics_s *pxys_stats;

struct msgoutQItem {
   size_t length;
   char   data[1040];
   struct msgoutQItem *  next;

};

/* ms_msgout special mutex */
pthread_mutex_t mutex_proxy_toast;
pthread_mutex_t mutex_send;

extern pthread_mutex_t mutex_link;

typedef struct msgoutQItem 	msgoutQItem;

/* Global: sendQ pointer */
static msgoutQItem *msgoutQ = NULL;
extern size_t msgoutSize;


void
msgout_init()
{
   /* allocate msgout queue array */
   msgoutQ = (msgoutQItem *) xcalloc(MAX_SENDQ_ITEMS, sizeof(msgoutQItem));
   
   pthread_mutex_init(&mutex_proxy_toast, NULL);
   pthread_mutex_init(&mutex_send, NULL);
}


void
msgout_free()
{
   /* delete */
   if (msgoutQ) {
      xfree(msgoutQ);
      pthread_mutex_destroy(&mutex_send);
      pthread_mutex_destroy(&mutex_proxy_toast);
   }
}

/* Messages to console
 * cached=0 -> real time scan
 * cached=1 -> proxy cached
 * cached=2 -> proxy cached and queued
 */

void
proxy_toast(struct in_addr *ip, unsigned int type, int cached)
{
   char ipaddr[16] = "000.000.000.000";
   char nn[6]; /* sizeof("YYXXX"); */
   struct Client *cptr;
   struct Relay_s *relay;
   long toasted;
   char *typeproxy;
   char *reason;
   char realtypeproxy[128];
   unsigned int port = type >> 16;
   type &= 0xFFFF;
   
   pthread_mutex_lock(&mutex_proxy_toast); /* too much code in that section :/ */
   
   get_proxy_type(type, port, cached, &typeproxy, &reason);
   
   if (!reason) {
      log_error("ms_msgout:proxy_toast: get_proxy_type() failed (%x)", type);
      return;
   }
   
   if (type == PROXY_TYPE_HTTPPROXY)
      snprintf(realtypeproxy, sizeof(realtypeproxy), "%s %u", typeproxy, port);
   else
      snprintf(realtypeproxy, sizeof(realtypeproxy), "%s", typeproxy);
   
   /* The string returned by inet_ntoa() resides in a static memory area. Just
    * to be sure, we're considering the next section as a critical one ;)
    */
   
   toasted = array_countclient_from_ip(ip); /* NB. mutex'd */
   
   strncpy(ipaddr, inet_ntoa(*ip), sizeof(ipaddr) - 1);
   
   if (!cached)
      totalfound++;
   
   /* 1- log - NB. log_write mutex'd */
   if (!cached) {
      log_write(LOGID_PROXY, "*@%s [%ld] %s", ipaddr, toasted, realtypeproxy);
   }
   
   /* 2- add to last log - NB. lastlog_put not mutex'd */
   lastlog_put(ipaddr, toasted, realtypeproxy);
   
   /* 3- record the gline for further auto-spread (see ms_glines.c) - NB. mutex'd */
   if (!cached) {
      time_t now;
      time(&now);
      proxycache_insert_new(ip, type, port, now + MyGline->cached);
   }
   
   /* 4- send to console channel - NB. will be queued and mutex'd */
   switch (cached) {
      case 0:
         client_sendmsgto_console("%s *@%s [%ld] %s", PROXY_GLINE_CMD,
                 ipaddr, toasted, realtypeproxy);
         break;
      case 1:
#ifdef CONSOLE_SHOWCACHED
         client_sendmsgto_console("%s *@%s [%ld] %s (cached)", PROXY_GLINE_CMD,
            ipaddr, toasted, realtypeproxy);
#endif
         break;
         
      case 2:
      /*
         client_sendmsgto_console("%s *@%s [%ld] %s (deque)", PROXY_GLINE_CMD,
            ipaddr, toasted, realtypeproxy) ;
       */
         break; /* was queued, print nothing */
   }
   
   /* 5- optionnaly, gline directly */
   if (MyGline->direct) {
      /* Log glines */
      log_write(LOGID_GLINES, "+*@%s %ld :[%ld] %s",
         ipaddr, MyGline->delay, toasted, reason);
      /* Send command */
#ifdef N2K_SUPPORT
      sendto_uplink("%s " TOK_GLINE " * +*@%s %ld :AUTO [%ld] %s", MyYY,
         ipaddr, MyGline->delay, toasted, reason);
#else
      sendto_uplink("%s " MSG_GLINE " * +*@%s %ld :AUTO [%ld] %s", MyYY,
         ipaddr, MyGline->delay, toasted, reason);
#endif
   }
   
   /* Using Uworld for glines? -> 2 checks: nick + resident server */
   
   if (MyGline->relay) {
      pthread_mutex_lock(&mutex_link);
      
      for (relay = MyGline->relay; relay; relay = relay->next) {
         if ((cptr = array_findclient_from_nick(relay->nick))) {
            if (hfind(mHS, relay->serv, strlen(relay->serv) + 1)) {
               if (((struct Server *)hstuff(mHS))->myx == cptr->nserv) {
                  
                  numnick(cptr, nn);
                  
                  if(*relay->auth && !relay->isauth) {
                     relay->isauth = 1;
#ifdef N2K_SUPPORT
                     sendto_uplink("%s%s %s %s :%s", MyYY, MyClientXXX,
                        TOK_PRIVATE, nn, relay->auth);
#else
                     sendto_uplink("%s%s %s %s :%s", MyYY, MyClientXXX,
                        MSG_PRIVATE, nn, relay->auth);
#endif
                  }
#ifdef N2K_SUPPORT
                  sendto_uplink("%s%s %s %s :" GLINE_SYNTAX, MyYY, MyClientXXX,
                     TOK_PRIVATE, nn, inet_ntoa(*ip), MyGline->delay, reason);
#else
                  sendto_uplink("%s%s %s %s :" GLINE_SYNTAX, MyYY, MyClientXXX,
                     MSG_PRIVATE, nn, inet_ntoa(*ip), MyGline->delay, reason);
#endif
                  break;
               }
            }
         }
      }
      pthread_mutex_unlock(&mutex_link);
   }
   pthread_mutex_unlock(&mutex_proxy_toast);
}


void
get_proxy_type(
   unsigned int type,
   unsigned int port,
   int cached,
   char **typeproxy,
   char **reason)
{
   switch (type) {
#ifdef SCAN_WINGATE
      case PROXY_TYPE_WINGATE:
         *typeproxy = PROXY_TYPESTR_WINGATE;
         *reason = MyGline->reason_wingate;
         if (!cached)
            pxys_stats->total_wingate++;
         break ;
#endif
#ifdef SCAN_SOCKS
      case PROXY_TYPE_SOCKS_V4:
         *typeproxy = PROXY_TYPESTR_SOCKS_V4;
         *reason = MyGline->reason_socks;
         if (!cached)
            pxys_stats->total_socks4++;
         break ;
      case PROXY_TYPE_SOCKS_V5:
         *typeproxy = PROXY_TYPESTR_SOCKS_V5;
         *reason = MyGline->reason_socks;
         if (!cached)
            pxys_stats->total_socks5++;
         break ;
#endif
#ifdef SCAN_HTTP
      case PROXY_TYPE_HTTPPROXY:
         *typeproxy = PROXY_TYPESTR_HTTPPROXY;
         *reason = MyGline->reason_http;
         if (!cached)
            stats_httpcount_addone(port);
         break ;
#endif
      default:
         *typeproxy = "Undefined proxy type";
         *reason = NULL;
         Debug((DBLMAX, "get_proxy_type: undefined type."));
         log_error("ms_msgout:get_proxy_type: undefined type (%x).", type);
         return;
   }
}


void
sendto_console(const char *pattern, ...)
{
   va_list vl;
   va_start(vl, pattern);
   vsendto_console(pattern, vl);
   va_end(vl);
}

void
vsendto_console(const char *pattern, va_list vl)
{
   char msgbuf[0x400], sendbuf[0x400];  /* local buf */

   vsnprintf(msgbuf, sizeof(msgbuf), pattern, vl);
#ifdef N2K_SUPPORT
   snprintf(sendbuf, sizeof(sendbuf), "%s %s %s :%s", MyYY, TOK_NOTICE, MyConsole->channel, msgbuf);
#else
   snprintf(sendbuf, sizeof(sendbuf), "%s %s %s :%s", MyYY, MSG_NOTICE, MyConsole->channel, msgbuf);
#endif

   send_buffer(sendbuf);
}

/* MyClient */

void
client_sendto_console(const char *pattern, ...)
{
   va_list vl;
   va_start(vl, pattern);
   vclient_sendto_console(pattern, vl);
   va_end(vl);
}

void
vclient_sendto_console(const char *pattern, va_list vl)
{
   char msgbuf[0x400], sendbuf[0x400];

   vsnprintf(msgbuf, sizeof(msgbuf), pattern, vl);
#ifdef N2K_SUPPORT
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s %s :%s", MyYY, MyClientXXX, TOK_NOTICE,
        MyConsole->channel, msgbuf);
#else
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s %s :%s", MyYY, MyClientXXX, MSG_NOTICE,
        MyConsole->channel, msgbuf);
#endif
   send_buffer(sendbuf);
}

void
client_sendmsgto_console(const char *pattern, ...)
{
   va_list vl;
   va_start(vl, pattern);
   vclient_sendmsgto_console(pattern, vl);
   va_end(vl);
}

void
vclient_sendmsgto_console(const char *pattern, va_list vl)
{
   char msgbuf[0x400], sendbuf[0x400];

   vsnprintf(msgbuf, sizeof(msgbuf), pattern, vl);
#ifdef N2K_SUPPORT
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s %s :%s", MyYY, MyClientXXX, TOK_PRIVATE,
        MyConsole->channel, msgbuf);
#else
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s %s :%s", MyYY, MyClientXXX, MSG_PRIVATE,
        MyConsole->channel, msgbuf);
#endif
   send_buffer(sendbuf);
}

/* MyClient --private */

void
client_sendto(const char *dst, const char *pattern, ...)
{
   va_list vl;
   va_start(vl, pattern);
   vclient_sendto(dst, pattern, vl);
   va_end(vl);
}

void
vclient_sendto(const char *dst, const char *pattern, va_list vl)
{
   char msgbuf[0x400], sendbuf[0x400];

   vsnprintf(msgbuf, sizeof(msgbuf), pattern, vl);
#ifdef N2K_SUPPORT
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s %s :%s", MyYY, MyClientXXX,
        TOK_NOTICE, dst, msgbuf);
#else
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s %s :%s", MyYY, MyClientXXX,
        MSG_NOTICE, dst, msgbuf);
#endif
   send_buffer(sendbuf);
}


/* SQUIT */

void
send_squit(const char *pattern, ...)
{
   va_list vl;
   va_start(vl, pattern);
   vsend_squit(pattern, vl);
   va_end(vl);
}

void
vsend_squit(const char *pattern, va_list vl)
{
   char msgbuf[0x400], sendbuf[0x400];  /* local buf */

   vsnprintf(msgbuf, sizeof(msgbuf), pattern, vl);

#ifdef N2K_SUPPORT
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s :%s", MyYY, MyClientXXX, TOK_QUIT, msgbuf);
#else
   snprintf(sendbuf, sizeof(sendbuf), "%s%s %s :%s", MyYY, MyClientXXX, MSG_QUIT, msgbuf);
#endif
   send_buffer(sendbuf);

#ifdef N2K_SUPPORT
   snprintf(sendbuf, sizeof(sendbuf), "%s %s %s 0 :%s", MyYY, TOK_SQUIT, Me->name, msgbuf);
#else
   snprintf(sendbuf, sizeof(sendbuf), "%s %s %s 0 :%s", MyYY, MSG_SQUIT, Me->name, msgbuf);
#endif
   send_buffer(sendbuf);
}


/* RAW messages to uplink (ms_auth,...) */

void
sendto_uplink(const char *pattern, ...)
{
   va_list vl;
   va_start(vl, pattern);
   vsendto_uplink(pattern, vl);
   va_end(vl);
}

void
vsendto_uplink(const char *pattern, va_list vl)
{
   char sendbuf[0x800] ;
	
   vsnprintf(sendbuf, sizeof(sendbuf), pattern, vl);
   send_buffer(sendbuf);
   //printf("[MSGOUT] %s", sendbuf);
}



/* send_buffer */

void
send_buffer(char *buf)
{
   size_t len = strlen(buf);
   if(buf[len - 1] != '\n') {
      if(len > 510)
         len = 510;
      buf[len++] = '\r';
      buf[len++] = '\n';
      buf[len] = '\0';
   }
   
   pthread_mutex_lock(&mutex_send); /* Don't queue several elements
                                            * at the same time :p
                                            */
   msgout_queue(buf, len);
   
   pthread_mutex_unlock(&mutex_send);
}


/* Pre-condition: msgout_init() must have been called once before! */
void
msgout_queue(const char* buf, size_t len)
{
   /* assert(msgoutQ != NULL); */
   
   if (msgoutSize == 0) {
      /*fall thru*/
   } else if (msgoutSize < MAX_SENDQ_ITEMS) {
      
      memmove(&msgoutQ[1], &msgoutQ[0], msgoutSize*sizeof(msgoutQItem));
      
   } else {
      /* Max msgoutQ Exceeded */
      log_error("Max msgoutQ exceeded!");
      m_shutdown();
      return;
   }
   
   memset(&msgoutQ[0], 0, sizeof(msgoutQItem));
   msgoutQ[0].length = min(len, sizeof(msgoutQ[0].data)-1);
   strncpy(msgoutQ[0].data, buf, msgoutQ[0].length);
   msgoutQ[0].next = NULL;
   msgoutSize++;
   
   Debug((DBLBASIC, "msgout_queue: len = %lu data = %s\n", msgoutQ[0].length, msgoutQ[0].data));
}


int
send_queued_now(int fd)
{
   for (; msgoutSize > 0; msgoutSize--) {
       int n = send(fd, msgoutQ[msgoutSize - 1].data, msgoutQ[msgoutSize - 1].length, 0);
       Debug((DBLBASIC, "send_queued_now: %s\n", msgoutQ[msgoutSize - 1].data));
       if(n < 0) {
          if(EWOULDBLOCK == errno || EAGAIN == errno || EINTR == errno)
             break;
          else
             return -1;
       }
   }
   return 0;
}

/* ms_msgout.c */
