/* $Id: pxys_sem.c,v 1.2 2002/11/15 19:50:57 mbuna Exp $
 */

#include "pxys_sem.h"

#include <assert.h>
#include <pthread.h>

#ifdef HAVE_MACH_MACH_H
#include <mach/mach.h>
#endif

void pxys_sem_init(pxys_sem_t *sem, int value)
{
#ifdef PXYS_SEM_MACH
  kern_return_t rv;
  rv = semaphore_create(mach_task_self(), &sem->sem, SYNC_POLICY_FIFO, value);
  assert(rv == KERN_SUCCESS);
#else
  pthread_mutex_init(&sem->mutex, NULL);
  pthread_cond_init(&sem->cond, NULL);
  sem->value = value;
#endif
}

void pxys_sem_destroy(pxys_sem_t *sem)
{
#ifdef PXYS_SEM_MACH
  kern_return_t rv;
  rv = semaphore_destroy(mach_task_self(), sem->sem);
  assert(rv == KERN_SUCCESS);
#else
  pthread_cond_destroy(&sem->cond);
  pthread_mutex_destroy(&sem->mutex);
#endif
}

void pxys_sem_P(pxys_sem_t *sem)
{
#ifdef PXYS_SEM_MACH
  semaphore_wait(sem->sem); /* might fail if pxys get killed (we don't care) */
#else
  pthread_mutex_lock(&sem->mutex);
  if(--sem->value < 0)
    pthread_cond_wait(&sem->cond, &sem->mutex);
  pthread_mutex_unlock(&sem->mutex);
#endif
}

int pxys_sem_timedP(pxys_sem_t *sem, pxys_sem_timespec_t *wait_time)
{
#ifdef PXYS_SEM_MACH
  return semaphore_timedwait(sem->sem, *wait_time);
#else
  int err = 0;
  
  pthread_mutex_lock(&sem->mutex);
  if(--sem->value < 0)
    err = pthread_cond_timedwait(&sem->cond, &sem->mutex, wait_time);
  pthread_mutex_unlock(&sem->mutex);
  return err;
#endif
}

void pxys_sem_V(pxys_sem_t *sem)
{
#ifdef PXYS_SEM_MACH
  semaphore_signal(sem->sem);
#else
  pthread_mutex_lock(&sem->mutex);
  if(++sem->value <= 0)
    pthread_cond_signal(&sem->cond);
  pthread_mutex_unlock(&sem->mutex);
#endif
}


